import mappy_rs
import mappy as mp
import pytest
from pathlib import Path

RESOURCES = (
    Path(__file__).parent.resolve().parent.resolve() / "resources/benchmarking"
)
FASTQ_PATH = RESOURCES / "fastq"
INDEX_PATH = RESOURCES / "index/hg38_simple.mmi"
_FILE_SUFFIXES = set([".fq", ".fastq", ".fastq.gz", ".fq.qz"])


def _gen_fastq(path: Path):
    """
    Generator returning fastq records from either a directory fo Fastq or a
    single Fastq file
    """
    if not path.exists():
        raise RuntimeError("Path doesn't exist")
    if path.is_dir():
        for f in path.iterdir():
            if set(f.suffixes).intersection(_FILE_SUFFIXES):
                yield from mp.fastx_read(str(f))
    else:
        if set(path.suffixes).intersection(_FILE_SUFFIXES):
            yield from mp.fastx_read(str(f))


N_READS = sum(1 for _ in _gen_fastq(FASTQ_PATH))


@pytest.fixture(scope="module")
def fasta():
    return [s for _, s, _ in _gen_fastq(FASTQ_PATH)]


@pytest.fixture
def mappy_al_rs():
    return mappy_rs.Aligner(str(INDEX_PATH))


@pytest.fixture
def mappy_al():
    yield mp.Aligner(str(INDEX_PATH))


@pytest.fixture
def aligner(request):
    yield request.getfixturevalue(request.param)


def align_multi(al, fasta):
    """
    Parameters
    ----------
    al: mappy_rs.Aligner
        Multithreaded aligner client
    fasta : list[str]
        List of sequences to align
    """
    res = al.map_batch({"seq": seq} for seq in fasta)
    return sum(1 for _ in res)


def _align(al, seqs):
    n = 0
    for s in seqs:
        _ = list(al.map(s))
        n += 1
    return n


@pytest.mark.parametrize("aligner", ["mappy_al", "mappy_al_rs"], indirect=True)
def test_classic_mappy(benchmark, aligner, fasta):
    n = benchmark.pedantic(
        _align, args=(aligner, fasta), iterations=1, rounds=1
    )
    assert N_READS == n
    print("Finished classic mappy round")


@pytest.mark.parametrize("threads", list(range(1, 6)))
def test_benchmark_multi(threads, benchmark, mappy_al_rs, fasta):
    mappy_al_rs.enable_threading(threads)
    n = benchmark.pedantic(
        align_multi, args=(mappy_al_rs, fasta), iterations=1, rounds=1
    )
    assert N_READS == n
    print("Finished threaded mappy round")



"""python_test.py

This set of tests roughly recreate those that are in lib.rs but use the
compiled package.
"""
from pathlib import Path
import copy

import pytest

import mappy_rs

RESOURCES = Path(__file__).parent.resolve().parent.resolve() / "resources/test"
MMI_FILE = RESOURCES / "test.mmi"
FA_FILE = RESOURCES / "test.fa"


def read_fasta(fh):
    for line in fh:
        if line.startswith(">"):
            name = line[1:].strip()
            break
    fa_lines = []
    for line in fh:
        if line.startswith(">"):
            yield name, "".join(fa_lines)
            fa_lines = []
            name = line[1:].strip()
            continue
        fa_lines.append(line.strip())
    yield name, "".join(fa_lines)


@pytest.fixture
def mmi_file():
    return str(MMI_FILE)


@pytest.fixture
def al(mmi_file):
    return mappy_rs.Aligner(mmi_file)


@pytest.fixture
def fasta(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def fasta_file():
    return str(FA_FILE)


@pytest.fixture
def fasta_list(fasta_file):
    with open(fasta_file, "rt") as fh:
        seqs = [s for _, s in read_fasta(fh)]

    seqs = [
        {"id": i, "seq": seq}
        for i, seq in enumerate(copy.copy(s) for _ in range(10) for s in seqs)
    ]
    return seqs


@pytest.fixture
def fasta_iter(fasta_list):
    return iter(fasta_list)


@pytest.fixture
def fasta_tuple(fasta_list):
    return tuple(fasta_list)


@pytest.fixture
def fasta_generator(fasta_list):
    return (item for item in fasta_list)


def test_test(al):
    assert al


def test_property_k(al):
    assert al.k == 15


def test_property_n_seq(al):
    assert al.n_seq == 4


def test_property_w(al):
    assert al.w == 10


def test_property_seq_names(al):
    expected = [
        "Bacillus_subtilis",
        "Enterococcus_faecalis",
        "Escherichia_coli_1",
        "Escherichia_coli_2",
    ]
    seq_names = al.seq_names
    seq_names.sort()
    assert seq_names == expected


def test_get_seq(al):
    contig = "Bacillus_subtilis"
    expected = (
        "AGAGTGAAGCCAATATTCCGATAACGATTGCTTTCATGATATCCCTCATTCTGGCATTATTTTTTTATA"
        "CTATACTATTCGATATCGCACAGATCAATGGAGTCGTGAGAAAATAAACATGTTTTGCGAACCGCTATG"
        "TGTGGAAGACAAAAAATGGAGGTGAAATTGATGGAAGCAAAGACACAGGCGTACTTTTTTCAGGATGAT"
        "GGCAGGATTCCGAATCACCCTGATTTTCCGCTCGTTGTGTATCAAAACGCACTCAAGGACACCGGTCAG"
        "GCAGAGCGGATCGTCAACCGGCATGGCTGGTCAAACAGCTGGTCGGGGAGTGTTTTTCCATACCATCAT"
        "TATCACAGCAATACGCATGAAGTCCTGATTGCAGTTCGGGGAGAGGCTGTGATTC"
    )
    seq = al.seq(contig)
    assert seq == expected


def test_map_one(al):
    mappings = al.map(
        "AGAGCAGGTAGGATCGTTGAAAAAAGAGTACTCAGGATTCCATTCAACTTTTACTGATTTGAAGCGTAC"
        "TGTTTATGGCCAAGAATATTTACGTCTTTACAACCAATACGCAAAAAAAGGTTCATTGAGTTTGGTTGT"
        "GATTTGATGAAAATTACTGAGAATAACAGGATTATTAAGCTGATTGATGAACTAAATCAGCTTAATAAA"
        "TATTCTTTGCAGATAGGAATATTTGGGGAAAATGATTCTTTTATGGCGATGTTGGCCCAAGTTCATGAA"
        "TTTGGGGTGACTATTCGTCCCAAAGGTCGTTTTCTTGTTATACCACTTATGAAAAAGTATAGAGGTAAA"
        "AGTCCACGTCAATTTGATTTGTTTTTTATGCAAACTAAAGAAAATCACAAGTTTT",
        cs=True,
    )
    assert len(mappings) == 1
    mapping = mappings[0]
    assert mapping.target_start == 0
    assert mapping.target_end == 400


@pytest.mark.parametrize(
    "fasta",
    ["fasta_iter", "fasta_list", "fasta_tuple", "fasta_generator"],
    indirect=True,
)
def test_map_batch(al, fasta):
    al.enable_threading(2)
    mappings = al.map_batch(fasta)
    n = 0
    for res in mappings:
        n += 1
    assert n == 40


def test_map_batch_fail_dict_single(al, fasta_iter):
    fasta = next(fasta_iter)
    al.enable_threading(2)
    with pytest.raises(TypeError) as excinfo:
        _ = al.map_batch(fasta)
    e = str(excinfo)
    assert "Unsupported batch type, pass a list, iter, generator or tuple" in e


def test_map_batch_fail_dict_many(al, fasta_iter):
    fasta = {i: dct for i, dct in enumerate(fasta_iter)}
    al.enable_threading(2)
    with pytest.raises(TypeError) as excinfo:
        _ = al.map_batch(fasta)
    e = str(excinfo)
    assert "Unsupported batch type, pass a list, iter, generator or tuple" in e


def test_map_batch_fail_list_str(al, fasta_iter):
    fasta = [dct["seq"] for dct in fasta_iter]
    al.enable_threading(2)
    with pytest.raises(TypeError) as excinfo:
        _ = al.map_batch(fasta)
    assert "Element in iterable is not a dictionary" in str(excinfo.value)


def test_map_batch_fail_no_seq_key(al, fasta_iter):
    fasta = [{"SEQ": dct["seq"]} for dct in fasta_iter]
    al.enable_threading(2)
    with pytest.raises(KeyError) as excinfo:
        _ = al.map_batch(fasta)
    assert "AHHH Key 🗝️  not found in iterated dictionary" in str(excinfo)


def test_map_batch_fail_seq_not_str(al, fasta_iter):
    fasta = [{"seq": dct["seq"].encode()} for dct in fasta_iter]
    al.enable_threading(2)
    with pytest.raises(ValueError) as excinfo:
        _ = al.map_batch(fasta)
    assert "`seq` must be a string" in str(excinfo)


def test_map_batch_fail_exhausted_iter(al, fasta_iter):
    _ = list(fasta_iter)
    al.enable_threading(2)
    mappings = al.map_batch(fasta_iter)
    assert len(list(mappings)) == 0



"""Run a test mapping infinitely the same reads infinitely"""
from mappy_rs import Aligner
import mappy as mp
from itertools import cycle, islice
from pathlib import Path
from pyfastx import Fastx
import time
import minimappers2


def yield_reads(
    fastq_path: Path = Path(
        "resources/benchmarking/fastq/fastq_runid_"
        "daf898996f80d549e26d8d49742e033decf51fdb_0.fastq.gz"
    ),
) -> str:
    """
    Return an infinite generator cycling through the same reads
    provided by the given fastq
    Parameters
    ----------
    fastq_path: Path
        The path to a fastq file
    Yields
    ------
    str
        Endless fasta formatted records from the fastq path
    """
    fx = Fastx(str(fastq_path.resolve()))
    fastas = [
        {"seq": seq, "Notes": "testing", "read_id": name}
        for name, seq, qual in fx
    ]
    yield from cycle(fastas)


def batched(iterable, n):
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


def single_threaded():
    """Single threaded batches of 200 reads in mappy rs"""
    al = Aligner("resources/chr20_hg38_22.mmi")
    # al.enable_threading(8)
    for i, fa_records in enumerate(
        batched(
            yield_reads(),
            200,
        )
    ):
        for record in fa_records:
            alignments = al.map(record["seq"])
            print(len(list(alignments)))


def mappy_align():
    """Align with mappy and test for memory"""
    al = mp.Aligner("resources/chr20_hg38_22.mmi")
    # al.enable_threading(8)
    for i, fa_records in enumerate(
        batched(
            yield_reads(Path()),
            200,
        )
    ):
        for record in fa_records:
            alignments = al.map(record["seq"])
            print(len(list(alignments)))


def mappy_rs_no_op_align():
    """Align returning dummy data rather calling out to minimap2-rs"""
    al = Aligner("resources/chr20_hg38_22.mmi")
    # al.enable_threading(8)
    for i, fa_records in enumerate(
        batched(
            yield_reads(Path()),
            200,
        )
    ):
        for record in fa_records:
            alignments = al.map_no_op(record["seq"])
            print(len(list(alignments)))


def mappy_rs_threaded_no_op_align():
    """Align multithreaded returning dummy
    data rather calling out to minimap2-rs"""
    al = Aligner("resources/chr20_hg38_22.mmi")
    al.enable_threading(8)
    for i, fa_records in enumerate(
        batched(
            yield_reads(Path()),
            200,
        )
    ):
        alignments = al.map_batch(fa_records)
        for alignment in alignments:
            del alignment


def map_with_minimappers2_single():
    """Including minimapper2 for completeness"""
    aligner = minimappers2.map_ont()
    aligner.index("resources/chr20_hg38_22.mmi")
    sequences = []
    for i, fa_records in enumerate(
        batched(
            yield_reads(Path()),
            200,
        )
    ):
        sequences = [
            minimappers2.Sequence(str(j), record["seq"])
            for (j, record) in enumerate(fa_records)
        ]
        alignments = aligner.map(sequences)
        print(alignments.shape)


def map_with_minimappers2_multi():
    """AHHHHHHH"""
    aligner = minimappers2.map_ont()
    aligner.threads(4)
    aligner.index("resources/chr20_hg38_22.mmi")
    sequences = []
    for i, fa_records in enumerate(
        batched(
            yield_reads(Path()),
            200,
        )
    ):
        sequences = [
            minimappers2.Sequence(str(j), record["seq"])
            for (j, record) in enumerate(fa_records)
        ]
        alignments = aligner.map(sequences)
        print(alignments.shape)


# @profile
def mappy_rs_threaded():
    al = Aligner("resources/chr20_hg38_22.mmi")
    al.enable_threading(8)
    for i, fa_records in enumerate(
        batched(
            yield_reads(),
            250,
        )
    ):
        alignments = al.map_batch(fa_records)
        for _alignment in alignments:
            continue
        time.sleep(0.1)
        # if i == 10:
        #     break

    # qprint(len(list(alignments)))


if __name__ == "__main__":
    # main()
    mappy_rs_threaded()
